import pandas as pd
import pyspark.sql.functions as func
from pyspark.sql.functions import row_number, monotonically_increasing_id, col
from pyspark.sql.window import Window

from cern.nxcals.pyquery.builders import *

from lhcsmapi.metadata.MappingMetadata import MappingMetadata
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class BusbarResistanceQuery(object):

    @staticmethod
    def query_current_mean_avg(system, devices, property_name, signal, t_start_nxcals, t_end_nxcals, translate_udf,
                               spark):
        # prepare NXCALS query
        i_meas = DevicePropertyQuery.builder(spark).system(system) \
            .startTime(t_start_nxcals).endTime(t_end_nxcals)

        # append device, property pairs
        for device in devices:
            i_meas = i_meas.entity().device(device).property(property_name)

        # build dataset
        i_meas_results = i_meas.buildDataset().select("acqStamp", signal, "device")

        # execute a query
        i_meas_results_ds = i_meas_results.withColumn("class", translate_udf(col("acqStamp"))).filter(
            col("class") != -1). \
            groupBy("device", "class").agg(func.mean(signal), func.stddev(signal)).collect()

        return pd.DataFrame(i_meas_results_ds,
                            columns=['I_MEAS_name', 'I_MEAS_class', 'I_MEAS_mean', 'I_MEAS_std']).sort_values(
            by=['I_MEAS_class'])

    @staticmethod
    def query_voltage_mean_avg(system, U_RES_names, t_start_nxcals, t_end_nxcals, translate_udf, spark):
        index = 0
        MAX_LIST_SIZE = 1000
        u_res_results_df_all = pd.DataFrame(columns=['U_RES_name', 'U_RES_class', 'U_RES_mean', 'U_RES_std', 'U_RES_max', 'U_RES_min'])
        while index < len(U_RES_names):
            u_res = VariableQuery.builder(spark).system(system) \
                .startTime(t_start_nxcals).endTime(t_end_nxcals)

            for U_RES_name in U_RES_names[index:index + MAX_LIST_SIZE]:
                u_res = u_res.variable(U_RES_name)

            u_res_results = u_res.buildDataset().select("nxcals_timestamp", 'nxcals_value', 'nxcals_variable_name')

            u_res_results_ds = u_res_results.withColumn("class", translate_udf(col("nxcals_timestamp"))).filter(
                col("class") != -1). \
                groupBy("nxcals_variable_name", "class").agg(func.mean('nxcals_value'),
                                                             func.stddev('nxcals_value'),
                                                             func.max('nxcals_value'),
                                                             func.min('nxcals_value')).collect()

            u_res_results_df_temp = pd.DataFrame(u_res_results_ds, columns=u_res_results_df_all.columns)
            u_res_results_df_all = pd.concat([u_res_results_df_all, u_res_results_df_temp], axis=0)

            index += MAX_LIST_SIZE

        return u_res_results_df_all

    @staticmethod
    def query_voltage_mean_avg_per_board_with_class(system, SIGNALS, t_start_nxcals, t_end_nxcals, translate_udf, spark,
                                                    board_type, prefix='U_RES'):
        shift = 0 if board_type == 'A' else 5

        index = 0
        MAX_LIST_SIZE = 1000
        u_res_results_df_all = pd.DataFrame(
            columns=['{}_name'.format(prefix), '{}_class'.format(prefix), '{}_mean'.format(prefix),
                     '{}_std'.format(prefix), '{}_count'.format(prefix)])
        while index < len(SIGNALS):
            u_res = VariableQuery.builder(spark).system(system) \
                .startTime(t_start_nxcals).endTime(t_end_nxcals)

            for U_RES_name in SIGNALS[index:index + MAX_LIST_SIZE]:
                u_res = u_res.variable(U_RES_name)

            u_res_results = u_res.buildDataset().select("nxcals_timestamp", 'nxcals_value', 'nxcals_variable_name')

            u_res_results_ds = u_res_results.withColumn("class", translate_udf(col("nxcals_timestamp"))).filter(
                col("class") != -1). \
                withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id()))).filter(
                (col("id") + shift) % 10 < 5). \
                groupBy("nxcals_variable_name", "class").agg(func.mean('nxcals_value'), func.stddev('nxcals_value'),
                                                             func.count('nxcals_value')).collect()

            u_res_results_df_temp = pd.DataFrame(u_res_results_ds, columns=u_res_results_df_all.columns)
            u_res_results_df_all = pd.concat([u_res_results_df_all, u_res_results_df_temp], axis=0)

            index += MAX_LIST_SIZE

        return u_res_results_df_all

    @staticmethod
    def get_pc_signal_system_property_devices(circuit_type):
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)

        I_MEAS = SignalMetadata.get_signal_name(circuit_type, circuit_names[0], 'PC', 'NXCALS', 'I_MEAS')
        I_MEAS_property = SignalMetadata.get_circuit_signal_database_metadata(circuit_type,
                                                                              circuit_names[0],
                                                                              'PC', 'NXCALS')['property']
        I_MEAS_system = SignalMetadata.get_circuit_signal_database_metadata(circuit_type,
                                                                            circuit_names[0], 'PC', 'NXCALS')['system']

        I_MEAS_devices = []
        for circuit_name in circuit_names:
            metadata_pc = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, 'PC',
                                                                              'NXCALS')
            I_MEAS_devices.append(metadata_pc['device'])

        return I_MEAS, I_MEAS_system, I_MEAS_property, I_MEAS_devices

    @staticmethod
    def get_bb_system_signals(circuit_type):
        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        metadata_bb = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_names[0], 'BUSBAR',
                                                                          'NXCALS')
        U_RES_system = metadata_bb['system']

        circuit_names = SignalMetadata.get_circuit_names(circuit_type)
        U_RES_names = MappingMetadata.get_busbars_for_circuit_name(circuit_type, circuit_names)
        U_RES_names = [f"{U_RES_name}:U_RES" for U_RES_name in U_RES_names]

        return U_RES_system, U_RES_names

    # ToDo: Duplication of code!
    # ToDo: This type of query shall be more generic!
    @staticmethod
    def query_voltage_mean_avg_per_board(system, U_RES_names, t_start_nxcals, t_end_nxcals, spark, board_type):
        shift = 0 if board_type == 'A' else 5

        index = 0
        MAX_LIST_SIZE = 1000
        u_res_results_df_all = pd.DataFrame(
            columns=['U_RES_name', 'U_RES_min', 'U_RES_max', 'U_RES_mean', 'U_RES_std', 'U_RES_count'])
        while index < len(U_RES_names):
            u_res = VariableQuery.builder(spark).system(system) \
                .startTime(t_start_nxcals).endTime(t_end_nxcals)

            for U_RES_name in U_RES_names[index:index + MAX_LIST_SIZE]:
                u_res = u_res.variable(U_RES_name)

            u_res_results = u_res.buildDataset().select("nxcals_timestamp", 'nxcals_value', 'nxcals_variable_name')

            u_res_results_ds = u_res_results. \
                withColumn('id', row_number().over(Window.orderBy(monotonically_increasing_id()))).filter(
                (col("id") + shift) % 10 == 0). \
                groupBy("nxcals_variable_name").agg(func.min('nxcals_value'), func.max('nxcals_value'),
                                                    func.mean('nxcals_value'), func.stddev('nxcals_value'),
                                                    func.count('nxcals_value')).collect()

            u_res_results_df_temp = pd.DataFrame(u_res_results_ds, columns=u_res_results_df_all.columns)
            u_res_results_df_all = pd.concat([u_res_results_df_all, u_res_results_df_temp], axis=0)

            index += MAX_LIST_SIZE

        return u_res_results_df_all
