import matplotlib.pyplot as plt
import math

from lhcsmapi.analysis.ee.EnergyExtractionAnalysis import EnergyExtractionAnalysis
from lhcsmapi.reference.Reference import Reference
from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.ExpSignalAnalysis import ExpSignalAnalysis


class EnergyExtractionPlot(object):

    @staticmethod
    def plot_ee_fgc(u_dump_res_1_sync_df, u_dump_res_2_sync_df, i_meas_df, timestamp_fgc, circuit_name):
        tau_u_dump_res_1 = ExpSignalAnalysis.calculate_char_time(u_dump_res_1_sync_df, t_start_decay=1, t_end_decay=120)
        tau_u_dump_res_2 = ExpSignalAnalysis.calculate_char_time(u_dump_res_2_sync_df, t_start_decay=1, t_end_decay=120)

        title = '{} {} EE: I_MEAS(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)

        ax1 = i_meas_df["I_MEAS"].plot(title=title, color='C0', legend=True, grid=True)
        ax1.title.set_size(20)
        ax1.set_ylabel("I_MEAS, [A]", color='C0', fontsize=15)
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.legend(loc='lower left')
        ax1.tick_params(labelsize=15)

        ax2 = ax1.twinx()
        u_dump_res_1_sync_df["U_DUMP_RES"].plot(ax=ax2, figsize=(15, 7.5), color='C2', legend=True)
        u_dump_res_2_sync_df["U_DUMP_RES"].plot(ax=ax2, color='C3', legend=True)
        ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

        plot_text = ("EE1: tau ={:.1f} s\nEE2: tau ={:.1f} s".format(tau_u_dump_res_1, tau_u_dump_res_2))
        ax2.annotate(plot_text, xytext=(max(u_dump_res_1_sync_df.index.max(), i_meas_df.index.max())*1.1, 0), xy=(0.6, 0), fontsize=15)
        ax2.legend(["U_DUMP_RES_1", "U_DUMP_RES_2"], loc='upper right')
        ax2.tick_params(labelsize=15)
        ax2.set_xlim((-50, ax2.get_xlim()[1]))
        plt.show()

    @staticmethod
    def plot_i_a_u_dump_res(i_a_df, i_ref_df, u_dump_res_1_sync_df, u_dump_res_2_sync_df, timestamp_fgc, timestamp_ee,
                            timestamp_pic, circuit_name):
        title = '{} {} EE: I_A(t) and U_DUMP_RES(t)'.format(Time.to_string_short(timestamp_fgc), circuit_name)
        ylim = (i_a_df['I_A'].max() - 100, i_a_df['I_A'].max() + 10)
        ax1 = i_a_df['I_A'].plot(title=title, xlim=(-0.2, 1), ylim=ylim, color='C0', legend=True, figsize=(15, 7.5),
                                 grid=True)
        i_ref_df['STATUS.I_REF'].plot(ax=ax1, color='C1', marker='o', style='-', xlim=(-0.2, 1), legend=True, grid=True)
        ax1.title.set_size(20)
        ax1.set_ylabel("I_A, [A]", color='C0', fontsize=15)
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.legend(loc='lower left')
        ax1.tick_params(labelsize=15)

        ax2 = ax1.twinx()
        u_dump_res_1_sync_df["U_DUMP_RES"].plot(ax=ax2, xlim=(-0.2, 1), color='C2', legend=True)
        u_dump_res_2_sync_df["U_DUMP_RES"].plot(ax=ax2, xlim=(-0.2, 1), color='C3', legend=True)
        ax2.set_ylabel("U_DUMP_RES, [V]", color='C2', fontsize=15)

        plot_text = ("EE1: delay = {:.1f} ms\nEE2: delay = {:.1f} ms".format((timestamp_ee[0] - timestamp_pic) * 1e-6,
                                                                         (timestamp_ee[1] - timestamp_pic) * 1e-6))
        ax2.annotate(plot_text, xytext=(1.1, 0), xy=(0.6, 0), fontsize=15)

        ax2.legend(["U_DUMP_RES_1", "U_DUMP_RES_2"], loc='upper right')

        ax2.annotate('', xy=((timestamp_ee[0] - timestamp_fgc) * 1e-9, 0.2 * ax2.get_ylim()[1]),
                     xytext=((timestamp_pic - timestamp_fgc) * 1e-9, 0.2 * ax2.get_ylim()[1]),
                     arrowprops=dict(arrowstyle='<->'))
        ax2.annotate('', xy=((timestamp_ee[1] - timestamp_fgc) * 1e-9, 0.1 * ax2.get_ylim()[1]),
                     xytext=((timestamp_pic - timestamp_fgc) * 1e-9, 0.1 * ax2.get_ylim()[1]),
                     arrowprops=dict(arrowstyle='<->'))
        ax2.tick_params(labelsize=15)

        plt.axvline(x=(timestamp_pic - timestamp_fgc) * 1e-9, color='green', linestyle='--', linewidth=1.5);
        plt.show()

    @staticmethod
    def plot_t_res_body(t_res_body_dfs, timestamp_fgc_act, circuit_name):
        # get reference temperature
        temp_ref_ee = Reference.get_energy_extraction_reference_temperature()

        title = '{} {} EE: T_RES_BODY versus time'.format(Time.to_string_short(timestamp_fgc_act), circuit_name)
        ax = t_res_body_dfs[0].plot(title=title, figsize=(15, 10))
        for t_res_body_df in t_res_body_dfs[1:]:
            t_res_body_df.plot(ax=ax)
        ax.grid(True)
        temp_ref_ee["Temp"].plot(ax=ax, color='C1', yerr=25, fmt='none', grid=True)
        ax.title.set_size(20)
        ax.set_xlabel("time, [s]", fontsize=15)
        ax.set_ylabel("T, [degC]", fontsize=15)
        ax.tick_params(labelsize=15)

        # get min_max_temperature
        t_res_body_max_min, t_res_body_max_max = EnergyExtractionAnalysis.get_min_max_t_res_body_max(t_res_body_dfs)

        plot_text = ('T_min_final = {} degC\nT_max_final = {} degC\n'.format(math.ceil(t_res_body_max_min),
                                                                             math.ceil(t_res_body_max_max)))
        ax.annotate(plot_text, xytext=(800, 0), xy=(700, 0), fontsize=15)

        plt.show()

    @staticmethod
    def plot_t_res_body_st_res_overtemp(t_res_body_dfs, st_res_overtemp_dfs, timestamp_fgc_act, circuit_name):
        title = '{} {} T_RES_BODY(t) and ST_RES_OVERTEMP(t)'.format(Time.to_string_short(timestamp_fgc_act),
                                                                    circuit_name)

        fig, ax1 = plt.subplots()
        fig.suptitle(title, fontsize=20)
        for t_res_body_df in t_res_body_dfs:
            t_res_body_df[t_res_body_df.index < 1000] = math.nan
            t_res_body_df.plot(ax=ax1, figsize=(15, 7.5), grid=True)
        ax1.legend(loc='lower left')
        ax1.set_xlabel("time, [s]", fontsize=15)
        ax1.set_ylabel("T_RES_BODY, [-]", fontsize=15)
        ax1.tick_params(labelsize=15)
        ax1.set_ylim([0, 200])
        ax1.axvspan(xmin=0, xmax=2*3600, ymin=0, ymax=0.35, facecolor='xkcd:yellowgreen')

        ax2 = ax1.twinx()
        ax2.set_ylabel("ST_RES_OVERTEMP, [-]", fontsize=15)
        ax2.tick_params(labelsize=15)
        for st_res_overtemp_df in st_res_overtemp_dfs:
            if not st_res_overtemp_df.empty:
                st_res_overtemp_df.plot(ax=ax2, drawstyle="steps-post")
        ax2.set_xlim([0, 3 * 3600])
        plt.show()

