import warnings
import math
import pandas as pd

from lhcsmapi.dbsignal.Signal import Signal
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


class CurrentLeadsQuery(object):


    @staticmethod
    def query_leads_events(circuit_type, circuit_name, timestamp_fgc, *, leads_type):
        # check if leads type is either LEADS_EVEN or LEADS_ODD
        if not any([leads_type == lt for lt in ['LEADS_EVEN', 'LEADS_ODD']]):
            raise ValueError('Leads type {} not in {}!'.format(leads_type, ['LEADS_EVEN', 'LEADS_ODD']))

        # Find PM events
        metadata_leads = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, leads_type,
                                                                             'PM')
        source_timestamp_leads = PmDbRequest.find_events(source=metadata_leads['source'],
                                                         system=metadata_leads['system'],
                                                         className=metadata_leads['className'],
                                                         t_start=timestamp_fgc, duration=[(10, 's'), (400, 's')])
        if len(source_timestamp_leads) != 1:
            warnings.warn('There should be only 1 PM entry, found: {}'.format(len(source_timestamp_leads)))
            return '', math.nan
        else:
            source_leads = source_timestamp_leads[0][0]
            timestamp_leads = source_timestamp_leads[0][1]
            return source_leads, timestamp_leads

    @staticmethod
    def query_u_leads(circuit_type, circuit_name, source_leads, timestamp_leads, *, leads_type):
        U_LEADS = SignalMetadata.get_signal_name(circuit_type, circuit_name, leads_type, 'PM', ['U_RES', 'U_HTS'])

        # check if leads type is either LEADS_EVEN or LEADS_ODD
        if not any([leads_type == lt for lt in ['LEADS_EVEN', 'LEADS_ODD']]):
            raise ValueError('Leads type {} not in {}!'.format(leads_type, ['LEADS_EVEN', 'LEADS_ODD']))

        if (source_leads is '') or math.isnan(timestamp_leads):
            warnings.warn('Either source_leads is empty or timestamp_leads is nan, returning an empty DataFrame')
            return pd.DataFrame(columns=[U_LEADS])

        metadata_leads = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, leads_type,
                                                                             'PM')

        u_leads_signal = Signal()
        for U_LEADS_el in U_LEADS:
            u_leads_signal.read('pm', append=True, source=source_leads, system=metadata_leads['system'],
                                className=metadata_leads['className'],
                                signal=U_LEADS_el, eventTime=timestamp_leads)

        u_leads_df = u_leads_signal.get_synchronized_time_df(timestamp_leads)

        return SignalUtilities.convert_indices_to_sec(u_leads_df)
