import sys
import os
from pathlib import Path

curr_dir = Path(os.path.split(os.getcwd())[0])
utilities_dir = str(curr_dir.parent / 'utilities')

if utilities_dir not in sys.path:
    sys.path.append(utilities_dir)

import numpy as np

import matplotlib.gridspec as gridspec

import matplotlib.pyplot as plt

# self made functions
import utils as u
import signalCALS as scals
import editDF as edf
import filterToolbox as ftb
import signalAnalysis as sa

# set global variables for history
log = []
warnings = []


class Dashboard():

    def enable_timestamp_plotting(resistance):
        """ This is necessary because of an error in the matplotlib.pyplot library """
        resistance['2010-03-14 12:14:14.14'].plot(style='o', markersize=3, figsize=(0.1, 0.1))

    def show_dash(time, location, big_signals, resistance, time_parameter, location_parameter, filter_text):
        """ Function creates Dashboard """
        # Dashboard-SETTINGS
        gs = gridspec.GridSpec(3, 15)
        gs.update(wspace=0, hspace=0.3)
        fig, plot = plt.subplots(figsize=(25, 15))
        ymargin = 1.5e-8  # for y_axis

        # TIME
        plt.subplot(gs[0, 0:8])
        k = Dashboard.time_plot(big_signals, location, ymargin)
        # TIME Histogram
        plt.subplot(gs[0, 8:10])
        mu_v, sigma_v = Dashboard.time_hist(big_signals, location, ymargin)

        # LOCATION
        plt.subplot(gs[1, 0:8])
        current_sector = Dashboard.location_plot(resistance, big_signals, time, location, ymargin)
        # LOCATION Histogram
        plt.subplot(gs[1, 8:10])
        mu_s, sigma_s = Dashboard.location_hist(current_sector, ymargin)

        # ORIGIN
        plt.subplot(gs[2, 0:8])
        snr, current_sn_rlow, current_sn_rhigh, percentage = Dashboard.origin_plot(resistance, big_signals, time,
                                                                                   time_parameter, location, ymargin)

        # History
        plt.subplot(gs[0:3, 10:13])
        # Generate Log-Text
        log_text = Dashboard.generate_log_text(time, location, big_signals, mu_v, sigma_v, mu_s, sigma_s, percentage)
        log.append(filter_text + log_text)
        warn_text = Dashboard.generate_warn_text(k, snr, current_sn_rlow, current_sn_rhigh)
        warnings.append(warn_text)
        Dashboard.history_plot(log, warnings)

    @staticmethod
    def time_plot(big_signals, location, ymargin):
        """ Function plots values over time """
        date = edf.EditDF.time(big_signals, 'date')  # Array with dates

        plt.ylabel('$R_{busbar} [\Omega]$')
        plt.xlabel('Time [years]')
        plt.title('$R_{busbar}  = f(t)$  @' + location)
        plt.ylim((0, ymargin))
        cmap = plt.get_cmap("winter")

        # plot all points from filtered Dataframe
        y = big_signals.values
        for i in range(0, len(y)):
            plt.plot(date, y[i], 'o', markersize=1, color=cmap(int(i * (256 / len(y)))))  # cmap has 256 colors

        # plot selected location in red
        loc = big_signals.index.get_loc(location)  # loc is index possition
        plt.plot(date, y[loc], 'o', markersize=2, color='r')

        # Linear regression of resitance
        reg, k = sa.SignalAnalysis.lin_reg(big_signals, location)

        plt.plot(date, reg, 'r--')
        plt.gcf().autofmt_xdate()

        plt.xticks(rotation='horizontal', ha='left')

        return k

    @staticmethod
    def time_hist(big_signals, location, ymargin):
        """ Function creates histogram of chosen location """
        v = big_signals.loc[location].values

        plt.title('Histogram Time')
        plt.hist(v, 100, range=[0, ymargin], align='mid', orientation='horizontal', color='r')

        v[np.isinf(v)] = 0
        mu_v = np.nanmean(v.astype('float64'))
        sigma_v = np.nanstd(v.astype('float64'))  # for Parameter-Log

        frame1 = plt.gca()
        plt.xticks([], [])
        frame1.axes.yaxis.set_ticklabels([])
        frame1.spines["top"].set_visible(False)
        frame1.spines["right"].set_visible(False)
        frame1.spines["bottom"].set_visible(False)
        return mu_v, sigma_v

    @staticmethod
    def location_plot(resistance, big_signals, time, location, ymargin):
        """ Function plots values over sectors """
        plt.ylim((0, ymargin))
        s = ['12', '23', '34', '45', '56', '67', '78', '81']
        # Memory allocation
        current_sector = {}
        ticks = np.zeros(len(s) + 1)
        ticks[0] = 0

        for i in range(len(s)):
            sector = ftb.FilterToolbox.sector(resistance, s[i])
            offset = 20
            ticks[i + 1] = ticks[i] + len(sector) + offset
            x = np.arange(ticks[i] + offset / 2, ticks[i] + offset / 2 + sector.shape[0])
            Y = sector[time].values
            # plot all
            plt.bar(x, Y, width=1, color='SkyBlue')
            for i in range(len(sector.index)):
                if sector.index[i] == location:
                    # plot sector
                    plt.bar(x, Y, width=1, color='b')
                    # plot chosen Signal
                    plt.bar(x[i], Y[i], width=3, color='r')
                    current_sector = Y
        plt.ylabel('$R_{busbar} [\Omega]$')
        plt.xlabel('Location [sector]')
        plt.title(r"$R_{busbar} = f(location)$  @" + time)

        xticklabels = np.arange(len(s) + 1)
        xticklabels[0] = 8
        plt.xticks(ticks, xticklabels)
        return current_sector

    @staticmethod
    def location_hist(current_sector, ymargin):
        """ Function creates histogram of sector in which the signal is in """
        plt.hist(current_sector, 100, range=[0, ymargin], density=1, align='mid', orientation='horizontal', color='b')
        mu_s = np.nanmean(current_sector)
        sigma_s = np.nanstd(current_sector)

        plt.title('Histogram Sector')
        frame1 = plt.gca()
        plt.xticks([], [])
        frame1.axes.yaxis.set_ticklabels([])
        frame1.spines["top"].set_visible(False)
        frame1.spines["right"].set_visible(False)
        frame1.spines["bottom"].set_visible(False)
        return mu_s, sigma_s

    @staticmethod
    def origin_plot(resistance, big_signals, time, time_parameter, location, ymargin):
        """ Function plots voltage and current which were taken to calculate the resistance """
        plt.title('$U_{RES}=f (t)$  $I_{MEAS}=f (t)$  @' + time + ' @' + location)
        plt.xlabel(r"Time [min]")
        plt.ylabel('Voltage [V]')
        # get full ramp, lower part of ramp and higher part of ramp
        (v_df, low_voltage, high_voltage) = sa.SignalAnalysis.get_signal_parts(time_parameter, time, location)
        v_index = v_df.index.values - v_df.index[0]  # index starts with 0

        plt.plot(v_index / 60, v_df.values, 'o', markersize=0.2)
        plt.plot(v_index[0:len(low_voltage.values)] / 60, low_voltage.values, 'bo', markersize=0.5)
        plt.plot(v_index[(len(v_index) - len(high_voltage.values)):len(v_index)] / 60, high_voltage.values, 'bo',
                 markersize=0.5)

        current_name = ftb.FilterToolbox.match_current(location)
        # get full ramp, lower part of ramp and higher part of ramp
        (i_df, low_current, high_current) = sa.SignalAnalysis.get_signal_parts(time_parameter, time, current_name)
        i_index = i_df.index.values - i_df.index[0]  # index starts with 0

        p = plt.twinx()
        p.plot(i_index / 60, i_df.values, 'o', markersize=1, color='orange')
        p.plot(i_index[0:len(low_current.values)] / 60, low_current.values, 'o', markersize=3, color='r')
        p.plot(i_index[(len(i_index) - len(high_current.values)):len(i_index)] / 60, high_current.values, 'o',
               markersize=3, color='r')
        p.set_ylabel('Current [A]')

        # Calculate parameters
        snr = sa.SignalAnalysis.snr(low_voltage.values)  # lower part of the voltage has lower SNR
        current_sn_rlow = sa.SignalAnalysis.snr(low_current.values)
        current_sn_rhigh = sa.SignalAnalysis.snr(high_current.values)
        percentage = sa.SignalAnalysis.percentage(len(big_signals.index.values), len(resistance.index.values))

        return (snr, current_sn_rlow, current_sn_rhigh, percentage)

    @staticmethod
    def history_plot(log, warnings):
        """ Function plots the history of the previous selections """
        # Creating figure and axis
        plt.gca().set_xlim(0., 1.)
        plt.gca().set_ylim(0., 1.)
        plt.gca().set_title("History")  # , weight='bold'
        plt.xticks([], [])
        plt.yticks([], [])
        # Write into Log
        n_lines = 3
        # Gap between lines in axes coords
        line_axesfrac = (1. / (n_lines))
        # shift, once the log is full
        if len(log) > n_lines:
            shift = len(log) - n_lines
        else:
            shift = 1
        for i_line in range(shift, n_lines + shift):
            baseline = 1 - (i_line - shift) * line_axesfrac
            baseline_next = baseline - line_axesfrac
            fill_color = ['grey', 'white'][i_line % 2]
            plt.fill_between([0., 1.], [baseline, baseline],
                             [baseline_next, baseline_next],
                             color=fill_color, alpha=0.15)
            if i_line < len(log):
                plt.annotate(log[i_line],
                             xy=(0.03, baseline - 0.8 * line_axesfrac),
                             xycoords='data', fontsize=9.5)
                plt.annotate(warnings[i_line],
                             xy=(0.03, baseline - 0.95 * line_axesfrac),
                             xycoords='data', color='r', fontsize=9.5)
            else:
                plt.annotate('',
                             xy=(0.07, baseline - 0.3 * line_axesfrac),
                             xycoords='data')

    @staticmethod
    def generate_log_text(time, location, big_signals, mu_v, sigma_v, mu_s, sigma_s, percentage):
        """ function generates the Log-Text """
        log_text = '\n' + '@' + time + ' @' + location \
                   + '\n' + "$R_{busbar}=$" + str(big_signals.loc[location][time]) + "$\Omega$" \
                   + '\n' + 'Histogram data:' \
                   + '\n' + '    Time: ' + r"$\mu$= " + str(mu_v)[0:5] + str(mu_v)[-4:len(
            str(mu_v))] + "$\Omega$   " + "$\sigma$=" + str(sigma_v)[0:5] + str(sigma_v)[
                                                                            -4:len(str(sigma_v))] + "$\Omega$" \
                   + '\n' + '    Sector: ' + r"$\mu$= " + str(mu_s)[0:5] + str(mu_s)[-4:len(
            str(mu_s))] + "$\Omega$   " + "$\sigma$=" + str(sigma_s)[0:5] + str(sigma_s)[-4:len(
            str(sigma_v))] + "$\Omega$" + '\n' + '\n'
        return log_text

    @staticmethod
    def generate_warn_text(k, snr, current_sn_rlow, current_sn_rhigh):
        """ function generates the Warning-Text"""
        warn_text = ''
        if k > 0 or k < -5e-10 or snr < 0.1 or current_sn_rlow < 100 or current_sn_rhigh < 100:
            warn_text = 'Warnings:'
        if k > 0:
            warn_text += '\n' + 'Derivative is positive: ' + str(k) + r" $\frac{ \Omega\ }{year}$"
        elif k < -5e-10:
            warn_text += '\n' + 'Splices could have been changed over time'
        if snr < 0.1:
            warn_text += '\n' + 'Voltage SNR is very small: ' + str(snr)
        if current_sn_rlow < 100 or current_sn_rhigh < 100:
            warn_text += '\n' + 'Inappropriate ramp for resistance calculation'
        return warn_text

    @staticmethod
    def list_to_txt(log, name):
        """ function writes list to txt """
        history = open(name + '.txt', 'w')
        for item in log:
            history.write("%s\n" % item)
