import json
import os
import re
import pandas as pd

from lhcsmapi.metadata.MappingMetadata import MappingMetadata


class SignalMetadata(object):
    BMODE = "HX:BMODE"

    __paths = {'RB':  'RB_METADATA.json',
               'RQ':  'RQ_METADATA.json',
               'IT': ' IT_METADATA.json',
               'IPD2_B1B2': 'IPD2_B1B2_METADATA.json',
               'IPD2': 'IPD2_METADATA.json',
               'IPQ2': 'IPQ2_METADATA.json',
               'IPQ4': 'IPQ4_METADATA.json',
               'IPQ8': 'IPQ4_METADATA.json',
               '600A': '600A_METADATA.json',
               '60A':  '60A_METADATA.json',
               '80-120A': '80-120A_METADATA.json'
               }

    @staticmethod
    def _parse_metadata(metadata_key):
        selected_path = os.path.dirname(__file__)
        try:
            selected_path = os.path.join(selected_path, SignalMetadata.__paths[metadata_key.upper()])
        except KeyError:
            raise KeyError("{} is not mapped to a json file".format(metadata_key))

        return SignalMetadata.__parse_json_file(selected_path)

    @staticmethod
    def get_beam_mode_details():
        # https://lhc-commissioning.web.cern.ch/lhc-commissioning/systems/data-exchange/published-topics/beam-modes.htm
        selected_path = os.path.dirname(__file__)
        full_path = os.path.join(selected_path, 'beam_mode/beam_mode.csv')
        return pd.read_csv(full_path, sep=';')

    @staticmethod
    def __parse_json_file(file_name):
        try:
            with open(file_name) as json_file:
                return json.load(json_file)
        except:
            raise FileNotFoundError("Could not load {}".format(file_name))

    @staticmethod
    def get_circuit_types():
        return list(SignalMetadata.__paths.keys())

    @staticmethod
    def get_circuit_names(circuit_type):
        circuit_types = circuit_type if type(circuit_type) is list else [circuit_type]
        circuit_names = []
        for circuit_type in circuit_types:
            circuit_names.extend(SignalMetadata._parse_metadata(circuit_type)['CIRCUITS'].keys())
        return circuit_names

    @staticmethod
    def get_system_types_per_circuit(circuit_type):
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        types = set()
        for circuit_name in circuit_metadata["CIRCUITS"].keys():
            if re.match(circuit_type, circuit_name):
                for key, value in circuit_metadata["CIRCUITS"][circuit_name].items():
                    if type(value) is dict:
                        types.add(value['TYPE'])
                    elif type(value) is list:
                        for element in value:
                            types.add(element['TYPE'])
        return list(types)

    @staticmethod
    def get_sector_names_from_components(circuit_type, system, *keys):
        if len(keys) == 1 and isinstance(keys[0], list):
            keys = keys[0]

        try:
            metadata = SignalMetadata._parse_metadata(circuit_type)['REL_IP_POS_TO_CIRCUIT_NAME']
            if any([system == sys for sys in ["MAGNET", "BUSBAR"]]):
                reg_exp = SignalMetadata._parse_metadata(circuit_type)[system]['REL_IP_POS_REG_EXP']
            elif system == "PC":
                reg_exp = SignalMetadata._parse_metadata(circuit_type)[system]['REL_IP_POS_REG_EXP']
                if len(keys) == 1:
                    return re.findall(reg_exp, keys[0])[0]
                else:
                    return [re.findall(reg_exp, key)[0] for key in keys]
            else:
                raise KeyError('System {} does not support this feature'.format(system))
            return SignalMetadata.__get_arc_names_from_magnet_names(metadata, reg_exp, *keys)
        except KeyError as e:
            raise KeyError(str(e) + "in object parsed from {}".format(SignalMetadata.__paths[circuit_type]))

    @staticmethod
    def __get_arc_names_from_magnet_names(metadata, reg_exp, *keys):
        if len(keys) == 1:
            return SignalMetadata.__get_sector_name_from_magnet_name(metadata, reg_exp, keys[0])
        else:
            return [SignalMetadata.__get_sector_name_from_magnet_name(metadata, reg_exp, key) for key in keys]

    @staticmethod
    def __get_sector_name_from_magnet_name(metadata, reg_exp, key):
        found_reg_exp = re.findall(reg_exp, key)
        if len(found_reg_exp) != 1:
            raise ValueError("Key {} does not contain proper MB magnet signal".format(key))
        else:
            key = found_reg_exp[0]
            if key not in metadata.keys():
                raise KeyError("Key {} does not exist ".format(key))
            else:
                return metadata[key]

    @staticmethod
    def get_circuit_name_metadata(circuit_type, circuit_name):
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)
        circuit_names = circuit_metadata["CIRCUITS"].keys()
        if circuit_name in circuit_names:
            return circuit_metadata["CIRCUITS"][circuit_name]
        else:
            raise KeyError("Circuit signal {} not present in {}.".format(circuit_name, circuit_names))

    @staticmethod
    def get_circuit_signal_database_metadata(circuit_type, circuit_name, system_type, database_name):
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)

        # Check if signal type is available
        system_types = SignalMetadata.get_system_types_per_circuit(circuit_type)
        if system_type in system_types:
            circuit_signal_type_metadata = circuit_metadata[system_type]
        else:
            raise KeyError("Signal type {} not present in {}.".format(system_type, system_types))

        # Check if databaseName is available
        database_names = circuit_signal_type_metadata.keys()
        if database_name in database_names:
            # update the source wildcard
            circuit_name_metadata = SignalMetadata.get_circuit_name_metadata(circuit_type, circuit_name)
            if database_name == "PM":
                source = circuit_signal_type_metadata[database_name]['source']
                if source in circuit_name_metadata.keys():
                    circuit_signal_type_metadata[database_name]['source'] = circuit_name_metadata[source]
                elif source in circuit_name_metadata[system_type].keys():
                    circuit_signal_type_metadata[database_name]['source'] = circuit_name_metadata[system_type][source]
            elif database_name == "NXCALS" and ('device' in circuit_signal_type_metadata[database_name]):
                device = circuit_signal_type_metadata[database_name]['device']
                if device in circuit_name_metadata.keys():
                    circuit_signal_type_metadata[database_name]['device'] = circuit_name_metadata[device]
                elif device in circuit_name_metadata[system_type].keys():
                    circuit_signal_type_metadata[database_name]['device'] = circuit_name_metadata[system_type][device]
            return circuit_signal_type_metadata[database_name]
        else:
            raise KeyError("Database signal {} not present in {}.".format(database_name, database_names))

    @staticmethod
    def get_signal_names(circuit_type, circuit_name, system_type, database_name):
        circuit_signal_type_db_metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type,
                                                                                                circuit_name,
                                                                                                system_type,
                                                                                                database_name)
        if database_name == "PM":
            output_list = list(circuit_signal_type_db_metadata.keys() - {"system", "className", "source"})
        else:
            output_list = list(circuit_signal_type_db_metadata.keys())

        output_list.sort()
        return output_list

    @staticmethod
    def get_signal_name(circuit_type, circuit_name, system_type, database_name, signal_name):
        if type(circuit_name) is not list:
            circuit_name = [circuit_name]

        signal_name = signal_name if type(signal_name) is list else [signal_name]

        signal_names = []
        for cn in circuit_name:
            for vn in signal_name:
                var_names = SignalMetadata.update_wild_cards_in_signal_name_per_circuit_name(cn, circuit_type,
                                                                                             database_name,
                                                                                             system_type, vn)
                if type(var_names) is list:
                    signal_names += var_names
                else:
                    signal_names.append(var_names)

        if len(signal_names) == 1:
            return signal_names[0]
        else:
            return signal_names

    @staticmethod
    def update_wild_cards_in_signal_name_per_circuit_name(circuit_name, circuit_type, database_name, system_type,
                                                          signal_name):
        circuit_signal_type_db_metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type,
                                                                                                circuit_name,
                                                                                                system_type,
                                                                                                database_name)
        signal_name = circuit_signal_type_db_metadata[signal_name]
        # if signal is a list
        if type(signal_name) is list:
            return [SignalMetadata.update_wild_cards_in_signal_name(circuit_name, circuit_type, system_type, vn) for vn in
                    signal_name]
        # if signal is a scalar
        else:
            return SignalMetadata.update_wild_cards_in_signal_name(circuit_name, circuit_type, system_type, signal_name)

    @staticmethod
    def update_wild_cards_in_signal_name(circuit_name, circuit_type, system_type, signal_name):
        # get all wildcards to be replaced
        circuit_name_metadata = SignalMetadata.get_circuit_name_metadata(circuit_type, circuit_name)
        # # for circuitName
        for key, value in circuit_name_metadata.items():
            if "%" in key:
                signal_name = signal_name.replace(key, value)
        # # for systemType
        system_type_metadata = circuit_name_metadata[system_type]
        for key, value in system_type_metadata.items():
            if key != "TYPE" and key in signal_name:
                if type(value) is list:
                    return [signal_name.replace(key, val) for val in value]
                else:
                    return signal_name.replace(key, value)
        return signal_name

    @staticmethod
    def get_database_names(circuit_type, system_type):
        circuit_metadata = SignalMetadata._parse_metadata(circuit_type)

        system_types = circuit_metadata.keys()
        if system_type in system_types:
            system_type_metadata = circuit_metadata[system_type]
        else:
            raise KeyError("System type {} not present in {}.".format(system_type, system_types))

        database_names = []
        for key, value in system_type_metadata.items():
            if any(db in key for db in ["PM", "CALS"]) and (len(value) > 0):
                database_names.append(key)

        return database_names

    @staticmethod
    def get_magnets_for_circuit_names(circuit_type, circuit_names):
        if type(circuit_names) is not list:
            circuit_names = [circuit_names]

        if circuit_type == "RB":
            magnet_to_circuit_name = MappingMetadata.get_magnet_cell_crate_circuit_name_mapping()
            are_magnets_matching_circuit_name = magnet_to_circuit_name["Sector"].isin(circuit_names)
            magnets_matching_circuit_name = magnet_to_circuit_name[are_magnets_matching_circuit_name]
            return list(magnets_matching_circuit_name['Magnet'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_cells_for_circuit_names(circuit_type, circuit_names):
        if type(circuit_names) is not list:
            circuit_names = [circuit_names]

        if circuit_type == "RB":
            magnet_to_circuit_name = MappingMetadata.get_magnet_cell_crate_circuit_name_mapping()
            are_magnets_matching_circuit_name = magnet_to_circuit_name["Sector"].isin(circuit_names)
            magnets_matching_circuit_name = magnet_to_circuit_name[are_magnets_matching_circuit_name]
            return list(magnets_matching_circuit_name['Cell'].values)
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_circuit_name_from_magnet_or_crate(circuit_type, magnet_or_crate):
        if circuit_type == 'RB':
            reg_exp = SignalMetadata._parse_metadata(circuit_type)['MAGNET']['REL_IP_POS_REG_EXP']
            suffix_to_sector_name = SignalMetadata._parse_metadata(circuit_type)['REL_IP_POS_TO_CIRCUIT_NAME']
            found_reg_exp = re.findall(reg_exp, magnet_or_crate)
            return suffix_to_sector_name[found_reg_exp[0]]
        else:
            raise KeyError('Circuit type: {} is not supported'.format(circuit_type))

    @staticmethod
    def get_wildcard_for_circuit_system_db_signal(circuit_type, circuit_name, system, db, signal):
        if any([system == sys for sys in ['QH', 'QDS']]) and circuit_type == 'RB':
            if db == "PM":
                wildcard = MappingMetadata.get_cells_for_circuit_names(circuit_type, circuit_name)
                return ('Cell', wildcard)
            elif db == 'CALS':
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                return ('Magnet', wildcard)
            else:
                raise KeyError("Database type {} not present in {}.".format(db, ['PM', 'CALS']))
        elif system == 'VF' and circuit_type == 'RB':
            wildcard = MappingMetadata.get_vf_for_circuit_names(circuit_type, circuit_name)
            return ('Voltage feeler', wildcard)
        elif system == "BUSBAR":
            wildcard = MappingMetadata.get_busbars_for_circuit_name(circuit_type, circuit_name)
            return ('Busbar', wildcard)
        elif system == 'DIODE_RB' and circuit_type == 'RB':
            if signal == 'U_REF_N1':
                wildcard = MappingMetadata.get_crates_for_circuit_names(circuit_type, circuit_name)
                return ('Crate', wildcard)
            else:
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                return ('Magnet', wildcard)
        elif system == 'DIODE_RQD' and circuit_type == 'RQ':
            if signal == 'U_REF_N1':
                wildcard = MappingMetadata.get_crates_for_circuit_names(circuit_type, circuit_name)
                return ('Crate', wildcard)
            else:
                wildcard = MappingMetadata.get_magnets_for_circuit_names(circuit_type, circuit_name)
                return ('Magnet', wildcard)
        return []

    @staticmethod
    def update_pm_source_with_wildcard(source_in, circuit, system, signal, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]

        if any([system == sys for sys in ['QH', 'QDS']]):
            return wildcard
        elif system == 'BUSBAR':
            return MappingMetadata.get_crate_name_from_busbar_names(circuit, wildcard)
        elif 'DIODE' in system:
            if signal == 'U_REF_N1':
                return wildcard
            else:
                return [MappingMetadata.get_crate_name_from_magnet_name(circuit, wc) for wc in wildcard]
        elif system == 'VF':
            return [MappingMetadata.get_crate_name_from_magnet_name(circuit, wc) for wc in wildcard]
        # for VF, PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return source_in

    @staticmethod
    def update_pm_signal_names_with_wildcard(signal_names_in, system, signal, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]
        if type(signal_names_in) is not list:
            signal_names_in = [signal_names_in]

        if system == 'QDS':
            return [signal_name.replace("%CELL%", wc) for wc in wildcard for signal_name in signal_names_in]
        elif system == 'BUSBAR':
            return [signal_name.replace("%BUSBAR%", wc) for wc in wildcard for signal_name in signal_names_in]
        elif 'DIODE' in system:
            if signal == 'U_REF_N1':
                return [signal_name.replace("%QPS_CRATE%", wc) for wc in wildcard for signal_name in signal_names_in]
            else:
                return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signal_names_in]
        elif system == 'VF':
            return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signal_names_in]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signal_names_in

    @staticmethod
    def update_cals_signal_names_with_wildcard(signal_names_in, system, wildcard):
        if type(wildcard) is not list:
            wildcard = [wildcard]
        if type(signal_names_in) is not list:
            signal_names_in = [signal_names_in]

        if any([system == sys for sys in ['QH', 'QDS', 'DIODE', 'VF']]):
            return [signal_name.replace("%MAGNET%", wc) for wc in wildcard for signal_name in signal_names_in]
        elif system == 'BUSBAR':
            return [signal_name.replace("%BUSBAR%", wc) for wc in wildcard for signal_name in signal_names_in]
        # for PC, EE, LEADS_EVEN, LEADS_ODD
        else:
            return signal_names_in
