import pandas as pd
import datetime
import re

from lhcsmapi.Time import Time
from lhcsmapi.dbsignal.Signal import Signal
from lhcsmapi.dbsignal.SignalUtilities import SignalUtilities
from lhcsmapi.dbsignal.post_mortem.PmDbRequest import PmDbRequest
from lhcsmapi.metadata.SignalMetadata import SignalMetadata


def validate_key_in_kwargs(key, *key_types, **kwargs):
    result = None
    if key in kwargs:
        result = kwargs[key]
        match_found = False
        for key_type in key_types:
            match_found = match_found or isinstance(result, key_type)
        if not match_found:
            raise TypeError("You entered a {} keyword, but the keyword needs to be a {}.".format(key, key_types))

    return result


def create_circuit_name_to_system_to_metadata_and_signal(circuit_type, circuit_name, system, db, signal, source):
    circuit_names = circuit_name if isinstance(circuit_name, list) else [circuit_name]
    systems = system if isinstance(system, list) else [system]

    map_circuit = {}
    for circuit_name in circuit_names:
        map_system = {}
        for system in systems:
            metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, db)
            signal_found = SignalMetadata.get_signal_name(circuit_type, circuit_name, system, db, signal)
            signal_found = signal_found if isinstance(signal_found, list) else [signal_found]
            metadata = {k: v for k, v in metadata.items() if k[0].islower()}
            if source is not None and db == 'PM':
                metadata['source'] = source
                signal_found = [s_f.replace('%CELL%', source) if '%CELL%' in s_f else s_f for s_f in signal_found]
                signal_found = [s_f.replace('%QPS_CRATE%', source) if '%QPS_CRATE%' in s_f else s_f for s_f in signal_found]

            map_system[system] = {'metadata': metadata, 'signal': signal_found}
        map_circuit[circuit_name] = map_system
    return map_circuit


def create_circuit_name_to_system_to_signal_to_params(circuit_name_to_system_to_metadata_and_signal, db, dbconnector):
    map_circuit_query = {}
    for circuit_name, system_metadata_signal in circuit_name_to_system_to_metadata_and_signal.items():
        map_system_query = {}
        for system, metadata_signal in system_metadata_signal.items():
            map_signal_query = {}
            for signal in metadata_signal['signal']:
                metadata = metadata_signal['metadata']
                metadata = update_metadata_for_db(db, metadata, signal, dbconnector)
                map_signal_query[signal] = metadata
            map_system_query[system] = map_signal_query
        map_circuit_query[circuit_name] = map_system_query

    return map_circuit_query


def update_metadata_for_db(db, metadata_db, signal, dbconnector):
    if db == 'PM':
        metadata_db = {k: v for k, v in metadata_db.items() if k[0].islower()}
    if db == 'NXCALS':
        metadata_db = {'nxcals_{}'.format(k): v for k, v in metadata_db.items() if k[0].islower()}
        metadata_db['spark'] = dbconnector
    elif db == 'CALS':
        metadata_db['ldb'] = dbconnector
    metadata_db['signal'] = signal

    return metadata_db


class Query(object):

    def set_db(self, db):
        self.db = db

    def set_circuit_type(self, circuit_type):
        self.circuit_type = circuit_type

    def set_t_start(self, t_start):
        self.t_start = t_start

    def set_t_end(self, t_end):
        self.t_end = t_end

    def set_timestamp(self, timestamp):
        self.timestamp = timestamp

    def set_signal(self, signal):
        self.signal = signal

    def set_circuit_name(self, circuit_name):
        self.circuit_name = circuit_name

    def set_system(self, system):
        self.system = system

    def set_metadata(self, metadata):
        self.metadata = metadata

    def set_circuit_name_to_system_to_metadata_and_signal(self, circuit_name_to_system_to_metadata_and_signal):
        self.circuit_name_to_system_to_metadata_and_signal = circuit_name_to_system_to_metadata_and_signal

    def get_time_params(self):
        time_params = {}
        if hasattr(self, 'timestamp') and self.db == 'PM':
            time_params['eventTime'] = self.timestamp
        elif self.db in ['NXCALS', 'CALS']:
            time_params['t_start'] = self.t_start
            time_params['t_end'] = self.t_end
        return time_params

    def get_signal_to_metadata_db(self, dbconnector):
        # Get metadata
        signals = [self.signal] if isinstance(self.signal, str) else self.signal
        metadata_db = {k: v for k, v in self.metadata.items() if k[0].islower()}

        # Correct keys and set additional metadata
        signal_to_metadata_db = {}
        for signal in signals:
            signal_to_metadata_db[signal] = update_metadata_for_db(self.db, metadata_db, signal, dbconnector)

        return signal_to_metadata_db


class QueryBuilder(object):
    """ Builder class to query databases (PM, CALS, NXCALS).
        The general syntax is
        QueryBuilder().with_db().with_circuit_type().{with_duration()/with_timestamp()}.with_metadata().{signal_query()/feature_query()}

        - EXAMPLE PM (event query)
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        source_timestamp_df = QueryBuilder().with_db('PM').with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end)\
                        .with_metadata(circuit_name=circuit_name, system='PC').event_query().df

        print(source_timestamp_df)
        >>>              source            timestamp
        >>> 0  RPTE.UA23.RB.A12  1421164827600000000

        - EXAMPLE PM (signal query)
        timestamp_fgc = 1421164827600000000
        source_fgc = 'RPTE.UA23.RB.A12'

        i_meas_df = QueryBuilder().with_db('PM').with_circuit_type('RB').with_timestamp(timestamp_fgc)\
                .with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query()\
                .synchronize_time(timestamp_fgc).convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>          STATUS.I_MEAS
        >>> -359.98      -0.150693
        >>> -359.96      -0.152328
        >>> -359.94      -0.151348
        >>> -359.92      -0.150366
        >>> -359.90      -0.149385

        - Example NXCALS (assuming connection to NXCALS cluster)::
        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_db('NXCALS').with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(spark).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>       I_MEAS
        >>> 0.0     0.07
        >>> 25.5    1.13
        >>> 26.0    1.72
        >>> 26.5    2.82
        >>> 27.0    4.32

        - EXAMPLE CALS

        import pytimber
        ldb = pytimber.LoggingDB()

        t_start = '2015-01-13 16:59:11+01:00'
        t_end = '2015-01-13 17:15:46+01:00'

        i_meas_df = QueryBuilder().with_db('CALS').with_circuit_type('RB').with_duration(t_start=t_start, t_end=t_end).\
                        with_metadata(circuit_name='RB.A12', system='PC', signal='I_MEAS').signal_query(ldb).synchronize_time().convert_index_to_sec().dfs[0]

        print(i_meas_df.head())
        >>>      RPTE.UA23.RB.A12:I_MEAS
        >>> 0.0                     0.07
        >>> 0.0                     1.13
        >>> 0.0                     1.72
        >>> 0.0                     2.82
        >>> 0.0                     4.32

        """

    def __init__(self):
        self.query = Query()

    def __getattr__(self, name):
        if name not in ['query', 'with_db']:
            raise AttributeError(
                '{}() is not supported, you can only call with_db(), e.g., with_db(\'NXCALS\')'.format(name))

    def __repr__(self):
        return 'Set database name using with_db() method, e.g., with_db(\'NXCALS\')'

    def with_db(self, db):
        # Check if db is supported
        self.query.set_db(db.upper())
        return QueryBuilderWithDb(self.query)


class QueryBuilderWithDb(object):

    def __init__(self, query):
        self.query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['query', 'with_circuit_type']:
            raise AttributeError(
                name + '() is not supported, you can only call with_circuit_type(), e.g., with_circuit_type(\'RB\')')

    def __repr__(self):
        return 'Database name properly set to {}. Set circuit type using with_circuit_type() method, e.g., with_circuit_type(\'RB\').'.format(
            self.query.db)

    def with_circuit_type(self, circuit_type):
        # Check if circuit_type is supported
        self.query.set_circuit_type(circuit_type.upper())
        return QueryBuilderWithDbCircuitType(self.query)


class QueryBuilderWithDbCircuitType(object):

    def __init__(self, query):
        self.query = query

    def __getattr__(self, name):
        if name not in ['query', 'with_duration']:
            raise AttributeError(
                '{}() not supported, you can call: \n with_duration(), e.g., with_duration(t_start=\'2015-01-13 16:59:11+01:00\', t_end=\'2015-01-13 17:15:46+01:00\')\n with_timestamp(), e.g. with_timestamp(1421164827600000000)'.format(
                    name))

    def __repr__(self):
        return 'Circuit type properly set to {}. Set duration using with_duration() method, e.g., with_duration(t_start=\'2015-01-13 16:59:11+01:00\', t_end=\'2015-01-13 17:15:46+01:00\').'.format(
            self.query.circuit_type)

    def with_timestamp(self, timestamp):
        if self.query.db == 'PM':
            self.query.set_timestamp(Time.to_unix_timestamp(timestamp))
            return QueryBuilderWithDbCircuitTypeTimeDef(self.query)
        else:
            raise ValueError('Timestamp can be only set for PM database, got {}'.format(self.query.db))

    def with_duration(self, **duration):
        t_start = validate_key_in_kwargs('t_start', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64,
                                         np.int64,
                                         **duration) or 0
        t_end = validate_key_in_kwargs('t_end', int, pd.Timestamp, str, float, datetime, tuple, np.datetime64, np.int64,
                                       **duration)
        duration = validate_key_in_kwargs('duration', tuple, list, **duration)

        t_start, t_end = Time.get_query_period_in_unix_time(t_start, t_end, duration, '')

        self.query.set_t_start(t_start)
        self.query.set_t_end(t_end)

        return QueryBuilderWithDbCircuitTypeTimeDef(self.query)


class QueryBuilderWithDbCircuitTypeTimeDef(object):

    def __init__(self, query):
        self.query = query

    # Can be generalized
    def __getattr__(self, name):
        if name not in ['query', 'with_metadata']:
            raise AttributeError(
                '{}() not supported, you can only call with_metadata(), e.g., with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\')'.format(
                    name))

    def __repr__(self):
        msg = 'Set metadata using with_metadata() method, e.g., with_metadata(circuit_name=\'RB.A12\', system=\'PC\', signal=\'I_MEAS\').'
        if hasattr(self.query, 'timestamp'):
            return 'Timestamp set to timestamp={}. {}'.format(self.query.timestamp, msg)
        else:
            return 'Query duration set to t_start={}, t_end={}. {}'.format(self.query.t_start, self.query.t_end, msg)

    # ToDo: with_source should be the next step of defining a query. This way, with_metadata does not become to complicated

    def with_metadata(self, **md):
        # Take the metadata according to date

        # Check if metadata definition is complete
        circuit_type = self.query.circuit_type
        db = self.query.db
        circuit_name = md['circuit_name']
        self.query.set_circuit_name(circuit_name)
        system = md['system']
        self.query.set_system(system)

        if 'source' in md:
            if isinstance(system, list) and (len(system) > 1):
                raise ValueError(
                    'source can only be set for a single system. For multiple systems, source is meaningless.')
            if isinstance(md['source'], list) and (len(md['source']) > 1):
                raise ValueError(
                    'source can only be a scalar. For multiple sources, query can lead to non-existing data in PM (mismatched timestamp).')
            source = md['source']
        else:
            source = None

        if 'signal' in md:
            circuit_name_to_system_to_metadata_and_signal = create_circuit_name_to_system_to_metadata_and_signal(
                circuit_type, circuit_name, system, db, md['signal'], source)

            # Check if in Signal there is a wildcard
            for circuit_name, system_to_metadata_signal in circuit_name_to_system_to_metadata_and_signal.items():
                for system, metadata_signal in system_to_metadata_signal.items():
                    signals = []
                    for signal in metadata_signal['signal']:
                        if '%' in signal:
                            # find wildcard in-between % signs
                            found_wildcard = re.findall('(?<=\%)(.*?)(?=\%)', signal)[0]

                            if 'wildcard' in md:
                                wildcard = md['wildcard']
                            else:
                                raise KeyError('In order to fully define a signal name ({}), please provide a wildcard as with_metadata(..., wildcard={{\'{}\': \'value\'}})'.format(signal, found_wildcard))

                            if found_wildcard in wildcard:
                                signal = signal.replace('%{}%'.format(found_wildcard), wildcard[found_wildcard])
                            else:
                                print(
                                    'Please provide a proper wildcard {0}, i.e. with_metadata(circuit_name=\'\', system=\'\', signal=\'\', wildcard={{\'{0}\': \'value\'}}'.format(
                                        found_wildcard))
                        signals.append(signal)
                    metadata_signal['signal'] = signals

            self.query.set_circuit_name_to_system_to_metadata_and_signal(circuit_name_to_system_to_metadata_and_signal)
        else:
            if isinstance(circuit_name, list):
                raise ValueError('Lists of circuit_name are not supported at the moment')
            metadata = SignalMetadata.get_circuit_signal_database_metadata(circuit_type, circuit_name, system, db)
            if '%' in metadata['source']:
                metadata['source'] = '*'
            if source is not None:
                metadata['source'] = source
            self.query.set_metadata(metadata)
        return QueryBuilderWithDbCircuitTypeTimeDefMetadata(self.query)


class QueryBuilderWithDbCircuitTypeTimeDefMetadata(object):

    def __init__(self, query):
        self.query = query

    # Can be generalized
    def __getattr__(self, name):
        if hasattr(self.query, 'timestamp') and name not in ['query', 'signal_query']:
            raise AttributeError('{}() not supported, for PM with timestamp you can call signal_query()'.format(name))
        elif hasattr(self.query, 't_start') and self.query.db == 'PM' and name not in ['query', 'event_query']:
            raise AttributeError('{}() not supported, for PM with duration you can call event_query()'.format(name))
        elif name not in ['query', 'signal_query']:
            raise AttributeError('{}() is not supported, you can call signal_query()'.format(name))

    def __repr__(self):
        if hasattr(self.query, 'timestamp'):
            return 'Query signal using signal_query() method, i.e., signal_query().'.format(
                self.query.timestamp)
        else:
            return 'Query signal using signal_query() method, i.e., signal_query().'.format(
                self.query.t_start, self.query.t_end)

    def signal_query(self, dbconnector=None):

        time_params = self.query.get_time_params()

        circuit_name_to_system_to_signal_to_params = create_circuit_name_to_system_to_signal_to_params(
            self.query.circuit_name_to_system_to_metadata_and_signal, self.query.db, dbconnector)

        circuit_dfs = []
        for circuit_name, system_to_signal_to_params in circuit_name_to_system_to_signal_to_params.items():
            system_dfs = []
            for system, signal_to_params in system_to_signal_to_params.items():
                signal_dfs = []
                for signal, params in signal_to_params.items():
                    query_params = {**time_params, **params}
                    signal_dfs.append(Signal().read(self.query.db, **query_params))
                system_dfs.append(signal_dfs)
            circuit_dfs.append(system_dfs)

        return SignalProcessing(circuit_dfs, self.query.db)

    def event_query(self):
        # This type of error only at the getattr level
        if self.query.db != 'PM':
            raise ValueError('Only PM database is supported, got {}.'.format(self.query.db))
        metadata = self.query.metadata
        t_start = self.query.t_start
        t_end = self.query.t_end

        # One can filter here for the circuit name
        event_source_timestamp = PmDbRequest.find_events_between_dates(t_start, t_end, metadata['className'],
                                                                       metadata['source'],
                                                                       metadata['system'])

        event_source_timestamp_df = pd.DataFrame(event_source_timestamp)
        event_source_timestamp_df.rename(columns={0: 'source', 1: 'timestamp'}, inplace=True)
        return EventProcessing(event_source_timestamp_df, self.query.system, self.query.circuit_name)

    def feature_query(self):
        pass


class EventProcessing(object):

    def __init__(self, df, system, circuit_name):
        self.df = df
        self.system = system
        self.circuit_name = circuit_name

    def filter_source(self):
        if self.system in ['QH', 'QDS']:
            if 'RQ' in self.circuit_name:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (SignalMetadata.get_sector_names_from_components(
                    'RQ', 'MAGNET', row['source'])[0] == self.circuit_name) |
                                                                     (SignalMetadata.get_sector_names_from_components(
                                                                         'RQ', 'MAGNET', row['source'])[
                                                                          1] == self.circuit_name), axis=1)
            else:
                self.df['is_circuit_ok'] = self.df.apply(lambda row: (
                        SignalMetadata.get_sector_names_from_components('RQ', 'MAGNET',
                                                                        row['source']) == self.circuit_name),
                                                         axis=1)

            self.df = self.df[self.df['is_circuit_ok']]
            self.df.drop(columns=['is_circuit_ok'], inplace=True)
        return self

    def drop_duplicate_source(self):
        self.df = self.df.drop_duplicates(subset=['source']).sort_values(
            by=['timestamp']).reset_index(drop=True)
        return self


import numpy as np


def flatten_nested_lists(x):
    result = []
    for el in x:
        if isinstance(el, (list, tuple)):
            result.extend(flatten_nested_lists(el))
        else:
            result.append(el)
    return result

class SignalProcessing(object):

    def __init__(self, dfs, db):
        self.dfs = flatten_nested_lists(dfs)
        self.db = db

    def __getattr__(self, name):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        if name not in method_list:
            raise AttributeError(name + '{}() is not supported, you can call {}'.format(name, method_list))

    def __repr__(self):
        method_list = [func for func in dir(SignalProcessing) if
                       callable(getattr(SignalProcessing, func)) and not func.startswith("__")]
        return 'Available methods are {}'.format(method_list)

    def flatten(self):
        self.dfs = flatten_nested_lists(self.dfs)
        return self

    def synchronize_time(self, t0=None):
        self.dfs = [SignalUtilities.synchronize_df(df, t0) for df in self.dfs]
        return self

    def convert_index_to_sec(self):
        self.dfs = [SignalUtilities.convert_indices_to_sec(df) for df in self.dfs]
        return self

    def create_col_from_index(self, col):
        for df in self.dfs:
            df[col] = df.index
        return self

    def filter_median(self, window=3):
        self.dfs = [df.rolling(window=window, min_periods=1).median() for df in self.dfs]
        return self

    def remove_values_for_time_less_than(self, time_threshold):
        self.dfs = [df[df.index > time_threshold] for df in self.dfs]
        return self

    def remove_initial_offset(self, time_threshold):
        self.dfs = [df.subtract(df[df.index < time_threshold].mean()) for df in self.dfs]
        return self
