import os
import math
import unittest
import json
import pandas as pd
from pathlib import Path

from lhcsmapi.analysis.qh.QuenchHeaterAnalysis import QuenchHeaterAnalysis
from lhcsmapi.analysis.qh.QuenchHeaterQuery import QuenchHeaterQuery
from lhcsmapi.dbsignal.SignalAnalysis import SignalAnalysis
from lhcsmapi.dbsignal.SignalElectrical import SignalElectrical
from lhcsmapi.metadata.SignalMetadata import SignalMetadata
from lhcsmapi.dbsignal.post_mortem.PmDbSignal import PmDbSignal
from test.test_dbsignal import test_ExponentialSignalAnalysis


class TestQuenchHeaterAnalysis(unittest.TestCase):

    @staticmethod
    def parsePmSignal(eventTime, source, signalName):
        system = "QPS"
        className = "DQAMCNMB_PMHSU"

        pms = PmDbSignal(system=system, className=className, source=source, signal=signalName, eventTime=eventTime)
        path = Path(os.path.dirname(__file__))
        file_path = os.path.join(str(path.parent.parent), 'resources/quench_heater_analysis/{}_{}_{}.json'.format(eventTime, source, signalName))
        with open(file_path) as json_file:
            response = json.load(json_file)
            content = response['content'][0]
            nameOfValue = pms._get_names_of_value(content)
            pms._get_time_value(nameOfValue, content)
            return pms._get_signal_df(signalName)

    def setUp(self):
        self.path = Path(os.path.dirname(__file__))

    @classmethod
    def setUpClass(self):
        circuit_type = 'RB'
        circuit_name = 'RB.A78'
        system_type = 'QH'

        U_HDS = SignalMetadata.get_signal_name(circuit_type, 'RB.A78', system_type, 'PM', 'U_HDS')
        I_HDS = SignalMetadata.get_signal_name(circuit_type, 'RB.A78', system_type, 'PM', 'I_HDS')
        R_HDS = [u.replace('U', 'R') for u in U_HDS]
        source = "A10L7"

        # Parse selected discharge
        timestamp = 1397227077614000000

        dfList = []
        for signalName in U_HDS + I_HDS:
            dfList.append(TestQuenchHeaterAnalysis.parsePmSignal(timestamp, source, signalName))

        self.df = pd.concat(dfList, axis=1)

        self.dfSync = QuenchHeaterQuery.preprocess_qh_discharge(circuit_type, circuit_name, self.df, timestamp)

        # Calculate resistance
        if circuit_type == 'RB':
            self.dfSync = SignalElectrical.calculate_resistances(self.dfSync, U_HDS, I_HDS, R_HDS)

        self.features = QuenchHeaterAnalysis.extract_features(self.dfSync, timestamp, U_HDS, I_HDS, R_HDS)

        # Parse reference discharge
        timestamp_ref = 1412844934913000000

        dfRefList = []
        for signalName in U_HDS + I_HDS:
            dfRefList.append(TestQuenchHeaterAnalysis.parsePmSignal(timestamp_ref, source, signalName))

        self.dfRef = pd.concat(dfRefList, axis=1)

        self.dfRefSync = QuenchHeaterQuery.preprocess_qh_discharge(circuit_type, circuit_name, self.dfRef, timestamp_ref)

        self.dfRefSync = SignalAnalysis.resample(self.dfRefSync)
        self.featuresRef = QuenchHeaterAnalysis.extract_features(self.dfRefSync, timestamp_ref, U_HDS, I_HDS, R_HDS)

        self.c2c = QuenchHeaterAnalysis.compare_curve_to_curve(self.dfSync, self.dfRefSync, self.featuresRef, I_HDS, R_HDS, U_HDS)

    def test_SelectedSignal_count(self):
        self.assertEqual(16384, self.df['U_HDS_1'].count())
        self.assertEqual(16384, self.df['U_HDS_2'].count())
        self.assertEqual(16384, self.df['U_HDS_3'].count())
        self.assertEqual(16384, self.df['U_HDS_4'].count())
        self.assertEqual(16384, self.df['I_HDS_1'].count())
        self.assertEqual(16384, self.df['I_HDS_2'].count())
        self.assertEqual(16384, self.df['I_HDS_3'].count())
        self.assertEqual(16384, self.df['I_HDS_4'].count())

    def test_SelectedSignal_min(self):
        self.assertAlmostEqual(0.0, self.df['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.df['I_HDS_4'].min(), places=5)

    def test_SelectedSignal_max(self):
        self.assertAlmostEqual(629.183100, self.df['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(1256.047100, self.df['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(125.001460, self.df['I_HDS_4'].max(), places=5)

    def test_SelectedSignal_mean(self):
        self.assertAlmostEqual(1.609766, self.df['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(2.375004, self.df['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(5.300451, self.df['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(4.584578, self.df['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(0.316847, self.df['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(0.293024, self.df['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(0.304800, self.df['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(0.193104, self.df['I_HDS_4'].mean(), places=5)

    def test_SelectedSignal_std(self):
        self.assertAlmostEqual(8.572087, self.df['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(32.265665, self.df['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(67.029452, self.df['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(61.729987, self.df['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(4.576926, self.df['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(4.237029, self.df['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(4.451355, self.df['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(2.167571, self.df['I_HDS_4'].std(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_count(self):
        self.assertEqual(16140, self.dfSync['U_HDS_1'].count())
        self.assertEqual(16140, self.dfSync['U_HDS_2'].count())
        self.assertEqual(16140, self.dfSync['U_HDS_3'].count())
        self.assertEqual(16140, self.dfSync['U_HDS_4'].count())
        self.assertEqual(16140, self.dfSync['I_HDS_1'].count())
        self.assertEqual(16140, self.dfSync['I_HDS_2'].count())
        self.assertEqual(16140, self.dfSync['I_HDS_3'].count())
        self.assertEqual(16140, self.dfSync['I_HDS_4'].count())

    def test_SelectedSignal_SyncDecayResampledSec_min(self):
        self.assertAlmostEqual(0.0575, self.dfSync['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0575, self.dfSync['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0575, self.dfSync['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0575, self.dfSync['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfSync['I_HDS_4'].min(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_max(self):
        self.assertAlmostEqual(98.187675, self.dfSync['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(1256.0471, self.dfSync['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(1256.0471, self.dfSync['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(1256.0471, self.dfSync['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(125.00146, self.dfSync['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(125.00146, self.dfSync['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(125.00146, self.dfSync['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(125.00146, self.dfSync['I_HDS_4'].max(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_mean(self):
        self.assertAlmostEqual(0.84591, self.dfSync['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(1.53169, self.dfSync['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(2.37237, self.dfSync['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(2.39363, self.dfSync['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(0.22569, self.dfSync['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(0.2023, self.dfSync['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(0.22272, self.dfSync['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(0.11024, self.dfSync['I_HDS_4'].mean(), places=5)

    def test_SelectedSignal_SyncDecayResampledSec_std(self):
        self.assertAlmostEqual(3.32749, self.dfSync['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(31.39285, self.dfSync['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(45.25308, self.dfSync['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(45.25237, self.dfSync['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(4.50393, self.dfSync['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(4.17561, self.dfSync['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(4.50403, self.dfSync['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(2.01865, self.dfSync['I_HDS_4'].std(), places=5)

    def test_SelectedSignal_CTimetau_charge(self):
        self.assertAlmostEqual(0.00326, self.features['U_HDS_1_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00315, self.features['U_HDS_2_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00032, self.features['U_HDS_3_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00032, self.features['U_HDS_4_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00029, self.features['I_HDS_1_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00248, self.features['I_HDS_2_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00028, self.features['I_HDS_3_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.00303, self.features['I_HDS_4_tau_charge'].values[0], places=5)

    def test_SelectedSignal_CTimetau_energy(self):
        self.assertAlmostEqual(0.00033, self.features['U_HDS_1_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.01667, self.features['U_HDS_2_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['U_HDS_3_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['U_HDS_4_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['I_HDS_1_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.01589, self.features['I_HDS_2_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.00021, self.features['I_HDS_3_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.00811, self.features['I_HDS_4_tau_energy'].values[0], places=5)

    def test_SelectedSignal_CTimeLinLog(self):
        self.assertAlmostEqual(-0.21782, self.features['U_HDS_1_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(-0.22514, self.features['U_HDS_2_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(-0.21798, self.features['U_HDS_3_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(-0.21595, self.features['U_HDS_4_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(774.59334, self.features['I_HDS_1_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(774.59334, self.features['I_HDS_2_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(774.59334, self.features['I_HDS_3_tau_lin_reg'].values[0], places=5)
        self.assertAlmostEqual(774.59334, self.features['I_HDS_4_tau_lin_reg'].values[0], places=5)

    def test_SelectedSignal_CTime_std(self):
        self.assertTrue(math.isnan(self.features['U_HDS_1_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_2_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_3_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['U_HDS_4_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_1_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_2_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_3_cTime_std'].values[0]))
        self.assertTrue(math.isnan(self.features['I_HDS_4_cTime_std'].values[0]))

    def test_SelectedSignal_CTime_mean(self):
        self.assertTrue(math.isinf(self.features['U_HDS_1_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_2_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_3_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['U_HDS_4_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_1_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_2_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_3_cTime_mean'].values[0]))
        self.assertTrue(math.isinf(self.features['I_HDS_4_cTime_mean'].values[0]))

    def test_SelectedSignal_Initial(self):
        self.assertAlmostEqual(89.19174, self.features['U_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(597.36039, self.features['U_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(1256.0471, self.features['U_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(1256.0471, self.features['U_HDS_4_initial'].values[0], places=5)
        self.assertAlmostEqual(125.00146, self.features['I_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(107.7168, self.features['I_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(125.00146, self.features['I_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(24.34515, self.features['I_HDS_4_initial'].values[0], places=5)

    def test_SelectedSignal_Final(self):
        self.assertAlmostEqual(0.84028, self.features['U_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(0.77875, self.features['U_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(0.85238, self.features['U_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(0.88164, self.features['U_HDS_4_final'].values[0], places=5)
        self.assertAlmostEqual(0.07439, self.features['I_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(0.07057, self.features['I_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(0.07198, self.features['I_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(0.0782, self.features['I_HDS_4_final'].values[0], places=5)

    def test_SelectedSignal_InitialResistance(self):
        self.assertAlmostEqual(0.71353, self.features['R_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(5.54566, self.features['R_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(10.04826, self.features['R_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(51.59332, self.features['R_HDS_4_initial'].values[0], places=5)

    def test_SelectedSignal_SimilarityMatrix(self):
        self.assertAlmostEqual(2.92646, self.features['U_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(4.25501, self.features['U_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(4.25501, self.features['U_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(3.00478, self.features['U_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(3.00479, self.features['U_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(0.00258, self.features['U_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(1.51701, self.features['I_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(0.00364, self.features['I_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(3.7169, self.features['I_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(1.517, self.features['I_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(3.35441, self.features['I_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(3.7169, self.features['I_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(-1.0, self.features['R_HDS_SimilarityMatrix_11'].values[0], places=5)

    def test_ReferenceSignal_count(self):
        self.assertEqual(16384, self.dfRef['U_HDS_1'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_2'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_3'].count())
        self.assertEqual(16384, self.dfRef['U_HDS_4'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_1'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_2'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_3'].count())
        self.assertEqual(16384, self.dfRef['I_HDS_4'].count())

    def test_ReferenceSignal_min(self):
        self.assertAlmostEqual(36.28133, self.dfRef['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(28.42325, self.dfRef['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(34.70972, self.dfRef['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(37.66129, self.dfRef['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(0.00954, self.dfRef['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(1.97034, self.dfRef['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(0.0, self.dfRef['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(0.02289, self.dfRef['I_HDS_4'].min(), places=5)

    def test_ReferenceSignal_max(self):
        self.assertAlmostEqual(989.63904, self.dfRef['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(935.0349, self.dfRef['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(949.352, self.dfRef['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(933.82745, self.dfRef['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(86.85918, self.dfRef['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(88.38892, self.dfRef['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(83.870285, self.dfRef['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(88.268745, self.dfRef['I_HDS_4'].max(), places=5)

    def test_ReferenceSignal_mean(self):
        self.assertAlmostEqual(441.29656, self.dfRef['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(421.29783, self.dfRef['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(444.61006, self.dfRef['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(440.4311, self.dfRef['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(34.9797, self.dfRef['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(33.7707, self.dfRef['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(34.11387, self.dfRef['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(34.53678, self.dfRef['I_HDS_4'].mean(), places=5)

    def test_ReferenceSignal_std(self):
        self.assertAlmostEqual(286.16068, self.dfRef['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(284.25813, self.dfRef['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(288.79218, self.dfRef['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(282.67392, self.dfRef['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(23.72157, self.dfRef['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(23.60303, self.dfRef['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(23.3162, self.dfRef['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(23.23312, self.dfRef['I_HDS_4'].std(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_count(self):
        self.assertEqual(63833, self.dfRefSync['U_HDS_1'].count())
        self.assertEqual(63833, self.dfRefSync['U_HDS_2'].count())
        self.assertEqual(63833, self.dfRefSync['U_HDS_3'].count())
        self.assertEqual(63833, self.dfRefSync['U_HDS_4'].count())
        self.assertEqual(63833, self.dfRefSync['I_HDS_1'].count())
        self.assertEqual(63833, self.dfRefSync['I_HDS_2'].count())
        self.assertEqual(63833, self.dfRefSync['I_HDS_3'].count())
        self.assertEqual(63833, self.dfRefSync['I_HDS_4'].count())

    def test_ReferenceSignal_SyncDecayResampledSec_min(self):
        self.assertAlmostEqual(36.28133, self.dfRefSync['U_HDS_1'].min(), places=5)
        self.assertAlmostEqual(28.44242, self.dfRefSync['U_HDS_2'].min(), places=5)
        self.assertAlmostEqual(34.72888, self.dfRefSync['U_HDS_3'].min(), places=5)
        self.assertAlmostEqual(37.66129, self.dfRefSync['U_HDS_4'].min(), places=5)
        self.assertAlmostEqual(2.61314, self.dfRefSync['I_HDS_1'].min(), places=5)
        self.assertAlmostEqual(1.97034, self.dfRefSync['I_HDS_2'].min(), places=5)
        self.assertAlmostEqual(2.40142, self.dfRefSync['I_HDS_3'].min(), places=5)
        self.assertAlmostEqual(2.69325, self.dfRefSync['I_HDS_4'].min(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_max(self):
        self.assertAlmostEqual(930.7226, self.dfRefSync['U_HDS_1'].max(), places=5)
        self.assertAlmostEqual(912.5723, self.dfRefSync['U_HDS_2'].max(), places=5)
        self.assertAlmostEqual(935.2458, self.dfRefSync['U_HDS_3'].max(), places=5)
        self.assertAlmostEqual(921.81036, self.dfRefSync['U_HDS_4'].max(), places=5)
        self.assertAlmostEqual(81.18657, self.dfRefSync['I_HDS_1'].max(), places=5)
        self.assertAlmostEqual(79.20669, self.dfRefSync['I_HDS_2'].max(), places=5)
        self.assertAlmostEqual(79.92197, self.dfRefSync['I_HDS_3'].max(), places=5)
        self.assertAlmostEqual(79.88763, self.dfRefSync['I_HDS_4'].max(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_mean(self):
        self.assertAlmostEqual(252.05261, self.dfRefSync['U_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(234.1204, self.dfRefSync['U_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(253.33099, self.dfRefSync['U_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(253.22598, self.dfRefSync['U_HDS_4'].mean(), places=5)
        self.assertAlmostEqual(19.81362, self.dfRefSync['I_HDS_1'].mean(), places=5)
        self.assertAlmostEqual(18.44683, self.dfRefSync['I_HDS_2'].mean(), places=5)
        self.assertAlmostEqual(19.42283, self.dfRefSync['I_HDS_3'].mean(), places=5)
        self.assertAlmostEqual(19.73932, self.dfRefSync['I_HDS_4'].mean(), places=5)

    def test_ReferenceSignal_SyncDecayResampledSec_std(self):
        self.assertAlmostEqual(228.20075, self.dfRefSync['U_HDS_1'].std(), places=5)
        self.assertAlmostEqual(224.73901, self.dfRefSync['U_HDS_2'].std(), places=5)
        self.assertAlmostEqual(231.01825, self.dfRefSync['U_HDS_3'].std(), places=5)
        self.assertAlmostEqual(226.14421, self.dfRefSync['U_HDS_4'].std(), places=5)
        self.assertAlmostEqual(18.46014, self.dfRefSync['I_HDS_1'].std(), places=5)
        self.assertAlmostEqual(18.20114, self.dfRefSync['I_HDS_2'].std(), places=5)
        self.assertAlmostEqual(18.24878, self.dfRefSync['I_HDS_3'].std(), places=5)
        self.assertAlmostEqual(18.14665, self.dfRefSync['I_HDS_4'].std(), places=5)

    def test_ReferenceSignal_CTimetau_charge(self):
        self.assertAlmostEqual(0.08994, self.featuresRef['U_HDS_1_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.08452, self.featuresRef['U_HDS_2_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.08979, self.featuresRef['U_HDS_3_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.09141, self.featuresRef['U_HDS_4_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.08068, self.featuresRef['I_HDS_1_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.07623, self.featuresRef['I_HDS_2_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.07997, self.featuresRef['I_HDS_3_tau_charge'].values[0], places=5)
        self.assertAlmostEqual(0.08237, self.featuresRef['I_HDS_4_tau_charge'].values[0], places=5)

    def test_ReferenceSignal_CTimeEnergyApproach(self):
        self.assertAlmostEqual(0.08532, self.featuresRef['U_HDS_1_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.0808, self.featuresRef['U_HDS_2_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.0859, self.featuresRef['U_HDS_3_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.08673, self.featuresRef['U_HDS_4_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.07142, self.featuresRef['I_HDS_1_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.06837, self.featuresRef['I_HDS_2_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.07104, self.featuresRef['I_HDS_3_tau_energy'].values[0], places=5)
        self.assertAlmostEqual(0.07329, self.featuresRef['I_HDS_4_tau_energy'].values[0], places=5)

    def test_ReferenceSignal_CTime_std(self):
        self.assertAlmostEqual(0.01011, self.featuresRef['U_HDS_1_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00917, self.featuresRef['U_HDS_2_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00855, self.featuresRef['U_HDS_3_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01016, self.featuresRef['U_HDS_4_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.0109, self.featuresRef['I_HDS_1_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00855, self.featuresRef['I_HDS_2_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.00953, self.featuresRef['I_HDS_3_cTime_std'].values[0], places=5)
        self.assertAlmostEqual(0.01106, self.featuresRef['I_HDS_4_cTime_std'].values[0], places=5)

    def test_ReferenceSignal_CTime_mean(self):
        self.assertAlmostEqual(0.09703, self.featuresRef['U_HDS_1_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09049, self.featuresRef['U_HDS_2_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09526, self.featuresRef['U_HDS_3_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09849, self.featuresRef['U_HDS_4_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09281, self.featuresRef['I_HDS_1_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.08562, self.featuresRef['I_HDS_2_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.09066, self.featuresRef['I_HDS_3_cTime_mean'].values[0], places=5)
        self.assertAlmostEqual(0.0942, self.featuresRef['I_HDS_4_cTime_mean'].values[0], places=5)

    def test_ReferenceSignal_Initial(self):
        self.assertAlmostEqual(930.21777, self.featuresRef['U_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(912.06173, self.featuresRef['U_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(934.71978, self.featuresRef['U_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(921.30148, self.featuresRef['U_HDS_4_initial'].values[0], places=5)
        self.assertAlmostEqual(81.01539, self.featuresRef['I_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(79.11819, self.featuresRef['I_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(79.48865, self.featuresRef['I_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(79.47032, self.featuresRef['I_HDS_4_initial'].values[0], places=5)

    def test_ReferenceSignal_Final(self):
        self.assertAlmostEqual(36.29714, self.featuresRef['U_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(28.45822, self.featuresRef['U_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(34.73594, self.featuresRef['U_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(37.68377, self.featuresRef['U_HDS_4_final'].values[0], places=5)
        self.assertAlmostEqual(2.61384, self.featuresRef['I_HDS_1_final'].values[0], places=5)
        self.assertAlmostEqual(1.97192, self.featuresRef['I_HDS_2_final'].values[0], places=5)
        self.assertAlmostEqual(2.40295, self.featuresRef['I_HDS_3_final'].values[0], places=5)
        self.assertAlmostEqual(2.69549, self.featuresRef['I_HDS_4_final'].values[0], places=5)

    def test_ReferenceSignal_InitialResistance(self):
        self.assertAlmostEqual(11.48199, self.featuresRef['R_HDS_1_initial'].values[0], places=5)
        self.assertAlmostEqual(11.52784, self.featuresRef['R_HDS_2_initial'].values[0], places=5)
        self.assertAlmostEqual(11.75916, self.featuresRef['R_HDS_3_initial'].values[0], places=5)
        self.assertAlmostEqual(11.59303, self.featuresRef['R_HDS_4_initial'].values[0], places=5)

    def test_ReferenceSignal_SimilarityMatrix(self):
        self.assertAlmostEqual(5.21185, self.featuresRef['U_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(0.9155, self.featuresRef['U_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(0.73976, self.featuresRef['U_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(5.79235, self.featuresRef['U_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(5.53086, self.featuresRef['U_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(1.37399, self.featuresRef['U_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(4.51208, self.featuresRef['I_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(1.43968, self.featuresRef['I_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(1.07016, self.featuresRef['I_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(3.26674, self.featuresRef['I_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(4.26744, self.featuresRef['I_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(1.07328, self.featuresRef['I_HDS_SimilarityMatrix_11'].values[0], places=5)

        self.assertAlmostEqual(13.23075, self.featuresRef['R_HDS_SimilarityMatrix_1'].values[0], places=5)
        self.assertAlmostEqual(31.02551, self.featuresRef['R_HDS_SimilarityMatrix_2'].values[0], places=5)
        self.assertAlmostEqual(9.0046, self.featuresRef['R_HDS_SimilarityMatrix_3'].values[0], places=5)
        self.assertAlmostEqual(27.46778, self.featuresRef['R_HDS_SimilarityMatrix_6'].values[0], places=5)
        self.assertAlmostEqual(13.69528, self.featuresRef['R_HDS_SimilarityMatrix_7'].values[0], places=5)
        self.assertAlmostEqual(22.26988, self.featuresRef['R_HDS_SimilarityMatrix_11'].values[0], places=5)

    def test_curve2curveComparison(self):
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['U_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=5)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.0, self.c2c['I_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.71171, self.c2c['R_HDS_1_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.69405, self.c2c['R_HDS_2_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.68129, self.c2c['R_HDS_3_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)
        self.assertAlmostEqual(0.54102, self.c2c['R_HDS_4_c2c_doubleTCEnvelope_globallynormalized'].values[0], places=4)

    def test_calculate_features_for_empty_df(self):
        # arrange
        circuit_type = 'RB'
        circuit_name = 'RB.A78'
        timestamp_qds = 1544269623912000000
        magnet = 'B29L8'
        qh_sync_decay_df = test_ExponentialSignalAnalysis.read_qh_discharge_from_csv(self.path.parent, timestamp_qds, magnet)

        timestamp_ref_qds = 1423589948613000000
        qh_ref_sync_decay_df = test_ExponentialSignalAnalysis.read_qh_discharge_from_csv(self.path.parent, timestamp_ref_qds, magnet)

        # act
        features_qh = QuenchHeaterAnalysis.calculate_actual_reference_features(circuit_type, circuit_name,
                                                                               qh_sync_decay_df, qh_ref_sync_decay_df)

        discharge_rating_act = QuenchHeaterAnalysis.rate_qh_discharge(circuit_type, features_qh)

        # assert
        discharge_rating_exp = False

        self.assertEqual(discharge_rating_exp, discharge_rating_act)